﻿'
' Copyright (c) 2021, MindFusion LLC - Bulgaria.
'

Imports System.ComponentModel
Imports System.Globalization
Imports System.Reflection

Imports Microsoft.Win32
Imports MindFusion.Diagramming.Wpf.Layout


Class MainWindow

	Public Sub New()

		InitializeComponent()

		overview1.Document = diagram
		overview1.ScaleMode = ScaleMode.CombinedScales
		diagram.Selection.Style = SelectionStyle.Frame
		_openFileDialog = New OpenFileDialog()

		diagram.TableRowHeight = 5 * coef
		diagram.TableColumnWidth = 4 * coef
		diagram.ShowAnchors = ShowAnchors.Never
		diagram.TableShape = SimpleShape.RoundedRectangle
		diagram.TableConnectionStyle = TableConnectionStyle.Both

		Dim tableNodeStyle As New Style()
		tableNodeStyle.Setters.Add(New Setter(TableNode.FontSizeProperty, 9.0))
		diagram.TableNodeStyle = tableNodeStyle
	End Sub

	Private coef As Double = 3.5


	Private Sub menuLoadAssembly_Click(ByVal sender As Object, ByVal e As RoutedEventArgs)
		If _openFileDialog.ShowDialog() = True Then
			' Create the TypeName <-> TableNode map
			_workMap = New Hashtable()

			' Create tables nodes and fill the tree
			ProcessAssembly(_openFileDialog.FileName)

			' Generate links between classes
			GenerateLinks()

			' Arrange the diagram
			Arrange()
		End If
	End Sub

	Private Sub menuExit_Click(ByVal sender As Object, ByVal e As RoutedEventArgs)
		Me.Close()
	End Sub

	Private Sub ProcessAssembly(ByVal assemblyFile As String)
		' Clear the tree
		treeView1.Items.Clear()

		' Clear diagram
		diagram.ClearAll()

		' Load the assembly
		Dim assembly__1 As Assembly

		Try
			assembly__1 = Assembly.LoadFrom(assemblyFile)
		Catch exception As Exception
			MessageBox.Show("Failed to load assembly." & vbLf & "Reason: """ + exception.Message & """", "Class Diagrammer", MessageBoxButton.OK, MessageBoxImage.[Error])
			Return
		End Try

		If assembly__1 Is Nothing Then
			MessageBox.Show("Failed to load assembly.", "ClassDiagrammer", MessageBoxButton.OK, MessageBoxImage.[Error])
			Return
		End If

		' Get exported types
		Dim types As Type() = assembly__1.GetExportedTypes()

		' Sort the exported types
		Dim sortedTypes As New ArrayList(types)
		sortedTypes.Sort(New TypeComparer())

		' Populate the tree view with assembly information
		For Each type As Type In sortedTypes
			' Add the namespace of the type
			Dim node = AddNamespace(type.[Namespace])

			' Add the type itself
			AddType(node, type)
		Next

		' Update tables' bounds
		For Each node As DiagramNode In diagram.Nodes
			Dim t As TableNode = TryCast(node, TableNode)
			If t Is Nothing Then
				Continue For
			End If

			Dim rows As Integer = Math.Min(8, t.RowCount)
			Dim height As Double = (7 + rows * 5) * coef

			t.Bounds = New Rect(t.Bounds.X, t.Bounds.Y, t.Bounds.Width, height)
		Next

		' Combine enums & delegates
		Dim enums As New DiagramNodeCollection()
		Dim delegates As New DiagramNodeCollection()
		For Each node As DiagramNode In diagram.Nodes
			Dim t As TableNode = TryCast(node, TableNode)
			If t Is Nothing Then
				Continue For
			End If

			Dim entity As Entity = TryCast(t.Tag, Entity)
			If entity.Type = ImageEnum Then
				enums.Add(t)
			ElseIf entity.Type = ImageDelegate Then
				delegates.Add(t)
			End If
		Next

		BuildGroup(enums, "Enums", ":enums")
		BuildGroup(delegates, "Delegates", ":delegates")
	End Sub

	Private Sub BuildGroup(ByVal tables As DiagramNodeCollection, ByVal name As String, ByVal tag As Object)
		If tables.Count > 0 Then
			Dim side As Integer = CInt(Math.Ceiling(Math.Sqrt(tables.Count)))
			Dim rows As Integer = CInt(Math.Ceiling(CDbl(tables.Count) / side))
			Dim singleWidth As Double = tables(0).Bounds.Width
			Dim width As Double = 5 * (side + 1) * coef + singleWidth * side
			Dim height As Double = (5 * (rows + 1) + 5) * coef

			Dim rowHeights As Double() = New Double(rows - 1) {}
			For i As Integer = 0 To tables.Count - 1
				rowHeights(i \ side) = Math.Max(tables(i).Bounds.Height, rowHeights(i \ side))
			Next

			For Each h As Double In rowHeights
				height += h
			Next

			'ShapeNode b = new ShapeNode(diagram);
			'b.Bounds = new Rect(0, 0, width, height);
			'ShapeNode b = diagram.Factory.CreateShapeNode(
			'	new Rect(0, 0, width, height));
			'diagram.Nodes.Add(b);
			Dim b As ContainerNode = diagram.Factory.CreateContainerNode(0, 0, width, height)

			Dim index As Integer = 0
			For Each e As TableNode In tables
				b.Add(e)

				Dim x As Double = (index Mod side) * (5 + singleWidth)
				Dim y As Double = 0

				For r As Integer = 0 To index \ side - 1
					y += 5 * coef + rowHeights(r)
				Next

				e.Bounds = New Rect(x + 5 * coef, y + 10 * coef, e.Bounds.Width, e.Bounds.Height)

				index += 1
			Next

			b.Caption = name
			b.TextAlignment = TextAlignment.Center
			b.TextVerticalAlignment = AlignmentY.Center
			b.HandlesStyle = HandlesStyle.HatchHandles3
			b.FontSize = 12
			b.FontWeight = FontWeights.Bold
			b.IgnoreLayout = True
			b.Tag = tag
			b.UpdateBounds()
		End If
	End Sub

	''' <summary>
	''' Walks the tree and creates the appropriate tables in the diagram.
	''' </summary>
	Private Sub GenerateLinks()
		' Create relation between types in the assembly
		For Each node As DiagramNode In diagram.Nodes
			Dim t As TableNode = TryCast(node, TableNode)
			If t Is Nothing Then
				Continue For
			End If

			'TreeViewItem treeNode = t.Tag as TreeViewItem;
			'Type type = treeNode.Tag as Type;
			Dim entity As Entity = TryCast(t.Tag, Entity)
			Dim type__1 As Type = TryCast(entity.Info, Type)
			Dim baseType As Type = type__1.BaseType
			If baseType Is Nothing Then
				Continue For
			End If

			Dim baseTypeName As String = baseType.[Namespace] + Type.Delimiter + baseType.Name

			Dim parent As TableNode = TryCast(_workMap(baseTypeName), TableNode)
			If parent IsNot Nothing Then
				'diagram.Links.Add(new DiagramLink(diagram, parent, t));
				diagram.Factory.CreateDiagramLink(parent, t)
			End If
		Next
	End Sub

	''' <summary>
	''' Arranges the diagram.
	''' </summary>
	Private Shadows Sub Arrange()
		Dim layout As New TreeLayout()

		layout.Type = TreeLayoutType.Centered
		layout.LinkStyle = TreeLayoutLinkType.Cascading3
		layout.LevelDistance = 25
		layout.KeepGroupLayout = True
		layout.Anchoring = Anchoring.Keep
		layout.MultipleGraphsPlacement = MultipleGraphsPlacement.MinimalArea
		layout.Margins = New Size(50, 50)

		layout.Arrange(diagram)

		' Place enums and delegates at the end
		Dim enums As DiagramNode = diagram.FindNode(":enums")
		Dim delegates As DiagramNode = diagram.FindNode(":delegates")

		Dim x As Double = 0
		For Each node As DiagramNode In diagram.Nodes
			Dim t As TableNode = TryCast(node, TableNode)
			If t Is Nothing Then
				Continue For
			End If

            If t.Container Is Nothing Then
                x = Math.Max(x, t.Bounds.Right)
            End If
		Next

		If enums IsNot Nothing Then
			enums.Move(x + 5, 5)
			x = enums.Bounds.Right
		End If

		If delegates IsNot Nothing Then
			delegates.Move(x + 5, 5)
		End If

		diagram.ResizeToFitItems(5)
	End Sub

	Private Function AddNamespace(ByVal namespaceName As String) As System.Windows.Controls.TreeViewItem
		Dim names As String() = namespaceName.Split(New Char() {"."c})

		Dim nodes As ItemCollection = treeView1.Items
		Dim node As System.Windows.Controls.TreeViewItem = Nothing

		For Each name As String In names
			node = Nothing
			For Each n2 As System.Windows.Controls.TreeViewItem In nodes
				If TryCast(n2.Header, Entity).Name = name Then
					node = n2
					Exit For
				End If
			Next

			If node Is Nothing Then
				' Find where to insert the new node
				Dim pos As Integer = nodes.Count
				For i As Integer = 0 To nodes.Count - 1
					Dim n3 = TryCast(nodes(i), System.Windows.Controls.TreeViewItem)
					If DefaultComparer.Compare(TryCast(n3.Header, Entity).Name, name) > 0 Then
						pos = i
					End If
				Next

				node = New System.Windows.Controls.TreeViewItem() With { _
				 .Header = New Entity() With { _
				  .Name = name, _
				  .Image = _ImageNamespace, _
				  .Info = Nothing, _
				  .Type = ImageNamespace _
				 } _
				}
				nodes.Insert(pos, node)
			End If

			nodes = node.Items
		Next

		Return node
	End Function

	Private Function AddType(ByVal parent As System.Windows.Controls.TreeViewItem, ByVal type As Type) As System.Windows.Controls.TreeViewItem
		If type.MemberType = MemberTypes.NestedType Then
			Dim stack As New Stack()

			Dim tempType As Type = type
			While tempType.MemberType = MemberTypes.NestedType
				tempType = tempType.DeclaringType
				stack.Push(tempType)
			End While

			Dim tempParent = parent
			While stack.Count > 0
				tempParent = AddTypeNoContents(tempParent, TryCast(stack.Pop(), Type))
			End While

			Return AddTypeWithContents(tempParent, type)
		Else
			Return AddTypeWithContents(parent, type)
		End If
	End Function

	''' <summary>
	''' Adds the specified type to the tree without
	''' adding the type contents.
	''' </summary>
	Private Function AddTypeNoContents(ByVal parent As System.Windows.Controls.TreeViewItem, ByVal type__1 As Type) As System.Windows.Controls.TreeViewItem
		' Do not add if already in
		For Each node As System.Windows.Controls.TreeViewItem In parent.Items
			If TryCast(node.Header, Entity).Name = type__1.Name Then
				Return node
			End If
		Next

		' Create a table for the type
		'TableNode t = new TableNode(diagram);
		Dim t As TableNode = diagram.Factory.CreateTableNode(New Rect(10 * coef, 10 * coef, 40 * coef, 10 * coef))
		t.CellTextStyle.FontSize = 9
		t.CellFrameStyle = CellFrameStyle.None
		t.TextAlignment = TextAlignment.Center
		t.TextVerticalAlignment = AlignmentY.Center
		'FormattedText form = new FormattedText(type.Name, CultureInfo.CurrentCulture, FlowDirection.LeftToRight, new Typeface(
		'    new FontFamily("Arial"),FontStyles.Normal,FontWeights.Bold,FontStretches.Normal), 15, Brushes.Black);
		t.Caption = type__1.Name
		t.RedimTable(3, 0)
		t.ToolTip = type__1.[Namespace] + Type.Delimiter + type__1.Name
		t.OffsetHeaderRows = True
		t.FontWeight = FontWeights.Bold
		'diagram.Nodes.Add(t);

		Dim entityType As Integer = -1
		Dim image As ImageSource = Nothing
		If type__1.IsClass Then
			If type__1.IsSubclassOf(GetType([Delegate])) Then
				entityType = ImageDelegate
				image = _ImageDelegate

				t.Brush = DelegateBrush
			Else
				entityType = ImageClass
				image = _ImageClass

				' Set brush and tooltip
				t.Brush = ClassBrush

				If Not type__1.BaseType.Equals(GetType([Object])) Then
					t.ToolTip += type__1.BaseType.[Namespace] + Type.Delimiter + type__1.BaseType.Name
				End If
			End If
		ElseIf type__1.IsInterface Then
			entityType = ImageInterface
			image = _ImageInterface

			t.Brush = InterfaceBrush
		ElseIf type__1.IsEnum Then
			entityType = ImageEnum
			image = _ImageEnum

			t.Brush = EnumBrush
		ElseIf type__1.IsValueType Then
			entityType = ImageStruct
			image = _ImageStruct

			t.Brush = StructBrush
		End If

		Dim child = New System.Windows.Controls.TreeViewItem()
		'child.Tag = type;
		child.Header = New Entity() With { _
		 .Name = type__1.Name, _
		 .Info = type__1, _
		 .Image = image, _
		 .Type = entityType _
		}
		parent.Items.Add(child)

		t.Tag = child.Header
		t.CustomDraw = CustomDraw.Additional
		t.AnchorPattern = AnchorPattern.TopInBottomOut

		_workMap(type__1.[Namespace] + Type.Delimiter + type__1.Name) = t

		Return child
	End Function

	''' <summary>
	''' Adds the specified type to the tree without recursing
	''' upwards if the type is nested.
	''' </summary>
	Private Function AddTypeWithContents(ByVal parent As System.Windows.Controls.TreeViewItem, ByVal type__1 As Type) As System.Windows.Controls.TreeViewItem
		' Add type item
		Dim child = AddTypeNoContents(parent, type__1)
		Dim entity As Entity = TryCast(child.Header, Entity)

		' Get the table that represents the type
		Dim table As TableNode = TryCast(_workMap(type__1.[Namespace] + Type.Delimiter + type__1.Name), TableNode)

		' Add contents
		'switch (parent.Items.IndexOf(child))
		Select Case entity.Type
			Case ImageClass, ImageStruct, ImageInterface
				If True Then
					' Enumerate methods first
					Dim methods As MethodInfo() = type__1.GetMethods(BindingFlags.Instance Or BindingFlags.[Public] Or BindingFlags.NonPublic Or BindingFlags.[Static] Or BindingFlags.DeclaredOnly)

					Dim constructors As ConstructorInfo() = type__1.GetConstructors(BindingFlags.Instance Or BindingFlags.[Public] Or BindingFlags.NonPublic Or BindingFlags.[Static] Or BindingFlags.DeclaredOnly)

					Dim sortedMethods As New ArrayList()
					sortedMethods.AddRange(methods)
					sortedMethods.AddRange(constructors)
					sortedMethods.Sort(New MemberInfoComparer())

					Dim methodsRow As Integer = -1

					For Each method As MethodBase In sortedMethods
						' We skip private stuff
						If method.IsPrivate Then
							Continue For
						End If

						' We skip internal stuff
						If method.IsAssembly Then
							Continue For
						End If

						' We process only methods here
						If method.IsSpecialName AndAlso Not method.IsConstructor Then
							Continue For
						End If

						Dim methodIcon As Integer = If(method.IsFamily, ImagePMethod, ImageMethod)

						Dim methodName As String = method.Name
						If method.MemberType = MemberTypes.Constructor Then
							methodName = method.DeclaringType.Name
						End If

						Dim childChild = New System.Windows.Controls.TreeViewItem()
						'childChild.Header = methodName;
						childChild.Header = New Entity() With { _
						 .Name = methodName, _
						 .Info = method, _
						 .Image = If(method.IsFamily, _ImagePMethod, _ImageMethod), _
						 .Type = If(method.IsFamily, ImagePMethod, ImageMethod) _
						}
						'childChild.Tag = method;
						child.Items.Add(childChild)

						' Add a row to the table (if not overload)
						Dim col As Integer = 2
						If Not method.IsStatic Then
							col = 1
						End If

						If methodsRow = -1 Then
							methodsRow = table.AddRow()
							table(0, methodsRow).ColumnSpan = 3
							table(0, methodsRow).Text = "Methods"
							table(0, methodsRow).TextAlignment = TextAlignment.Left
							table(0, methodsRow).TextVerticalAlignment = AlignmentY.Center
							table(0, methodsRow).FontWeight = FontWeights.Bold
							table.Rows(methodsRow).Header = True
							table.Rows(methodsRow).Expanded = True
						Else
							Dim found As Boolean = False
							For r As Integer = methodsRow + 1 To table.RowCount - 1
								If table(col, r).Text = methodName Then
									Dim [overloads] As Integer = CInt(table(col, r).Tag) + 1
									table(col, r).ToolTip = methodName & " (+ " & [overloads] & (If([overloads] = 1, " overload)", " overloads)"))
									table(col, r).Tag = [overloads]

									found = True
									Exit For
								End If
							Next

							If found Then
								Continue For
							End If
						End If

						Dim row As Integer = table.AddRow()
						If Not method.IsStatic Then
							table(col, row).ColumnSpan = 2
						End If

						table(0, row).Image = DirectCast(Application.Current.Resources("Method"), ImageSource)
						table(0, row).ImageAlign = ImageAlign.Center
						table(col, row).Text = methodName
						table(col, row).TextAlignment = TextAlignment.Left
						table(col, row).TextVerticalAlignment = AlignmentY.Center
						table(col, row).ToolTip = methodName
						table(col, row).Tag = 0

						If method.IsStatic Then
							table(1, row).Image = DirectCast(Application.Current.Resources("Static"), ImageSource)
							table(1, row).ImageAlign = ImageAlign.Center
						End If
					Next

					' Enumerate properties
					Dim properties As PropertyInfo() = type__1.GetProperties(BindingFlags.Instance Or BindingFlags.[Public] Or BindingFlags.NonPublic Or BindingFlags.[Static] Or BindingFlags.DeclaredOnly)

					Dim sortedProperties As New ArrayList(properties)
					sortedProperties.Sort(New MemberInfoComparer())

					Dim propertiesRow As Integer = -1

					For Each [property] As PropertyInfo In sortedProperties
						Dim getMethod As MethodInfo = [property].GetGetMethod(True)
						Dim setMethod As MethodInfo = [property].GetSetMethod(True)

						If getMethod Is Nothing AndAlso setMethod Is Nothing Then
							Continue For
						End If

						Dim isPublic As Boolean = False
						Dim isProtected As Boolean = False
						Dim isStatic As Boolean = False

						If getMethod IsNot Nothing Then
							isPublic = isPublic OrElse getMethod.IsPublic
							isProtected = isPublic OrElse getMethod.IsFamily
							isStatic = isStatic OrElse getMethod.IsStatic
						End If

						If setMethod IsNot Nothing Then
							isPublic = isPublic OrElse setMethod.IsPublic
							isProtected = isPublic OrElse setMethod.IsFamily
							isStatic = isStatic OrElse setMethod.IsStatic
						End If

						If Not isPublic AndAlso Not isProtected Then
							Continue For
						End If

						Dim propertyIcon As Integer = If(isPublic, ImageProperty, ImagePProperty)

						Dim childChild = New System.Windows.Controls.TreeViewItem()
						'childChild.Header = property.Name;
						'childChild.Tag = property;
						childChild.Header = New Entity() With { _
						 .Name = [property].Name, _
						 .Info = [property], _
						 .Image = If(isPublic, _ImageProperty, _ImagePProperty), _
						 .Type = If(isPublic, ImageProperty, ImagePProperty) _
						}
						child.Items.Add(childChild)

						' Add a row to the table (if not overload)
						Dim col As Integer = 2
						If Not isStatic Then
							col = 1
						End If

						If propertiesRow = -1 Then
							propertiesRow = table.AddRow()
							table(0, propertiesRow).ColumnSpan = 3
							table(0, propertiesRow).Text = "Properties"
							table(0, propertiesRow).TextAlignment = TextAlignment.Left
							table(0, propertiesRow).TextVerticalAlignment = AlignmentY.Center
							table(0, propertiesRow).FontWeight = FontWeights.Bold
							table.Rows(propertiesRow).Header = True
							table.Rows(propertiesRow).Expanded = True
						End If

						Dim description As String = ""
						Dim attributes As Object() = Nothing
						Try
							attributes = [property].GetCustomAttributes(False)
						Catch
							attributes = New Object(-1) {}
						End Try
						For Each attribute As Attribute In attributes
							If TypeOf attribute Is DescriptionAttribute Then
								description = TryCast(attribute, DescriptionAttribute).Description
								Exit For
							End If
						Next

						Dim row As Integer = table.AddRow()
						If Not isStatic Then
							table(col, row).ColumnSpan = 2
						End If

						table(0, row).Image = DirectCast(Application.Current.Resources("Property"), ImageSource)
						table(0, row).ImageAlign = ImageAlign.Center
						table(col, row).Text = [property].Name + " : " + [property].PropertyType.Name
						table(col, row).TextAlignment = TextAlignment.Left
						table(col, row).TextVerticalAlignment = AlignmentY.Center
						table(col, row).ToolTip = [property].Name + " : " + [property].PropertyType.[Namespace] + Type.Delimiter + [property].PropertyType.Name
						If description.Length > 0 Then
							table(col, row).ToolTip += description
						End If

						If isStatic Then
							table(1, row).Image = DirectCast(Application.Current.Resources("Static"), ImageSource)
							table(1, row).ImageAlign = ImageAlign.Center
						End If
					Next

					' Enumerate events
					Dim events As EventInfo() = type__1.GetEvents(BindingFlags.Instance Or BindingFlags.[Public] Or BindingFlags.NonPublic Or BindingFlags.DeclaredOnly)

					Dim sortedEvents As New ArrayList(events)
					sortedEvents.Sort(New MemberInfoComparer())

					Dim eventsRow As Integer = -1

					For Each eventInfo As EventInfo In sortedEvents
						Dim addMethod As MethodInfo = eventInfo.GetAddMethod(True)
						Dim removeMethod As MethodInfo = eventInfo.GetRemoveMethod(True)

						If addMethod Is Nothing AndAlso removeMethod Is Nothing Then
							Continue For
						End If

						Dim isPublic As Boolean = False
						Dim isStatic As Boolean = False

						If addMethod IsNot Nothing Then
							isPublic = isPublic OrElse addMethod.IsPublic
							isStatic = isStatic OrElse addMethod.IsStatic
						End If

						If removeMethod IsNot Nothing Then
							isPublic = isPublic OrElse removeMethod.IsPublic
							isStatic = isStatic OrElse removeMethod.IsStatic
						End If

						If Not isPublic Then
							Continue For
						End If

						Dim childChild = New System.Windows.Controls.TreeViewItem()
						'childChild.Header = eventInfo.Name;
						'childChild.Tag = eventInfo;
						childChild.Header = New Entity() With { _
						 .Name = eventInfo.Name, _
						 .Info = eventInfo, _
						 .Image = _ImageEvent, _
						 .Type = ImageEvent _
						}
						child.Items.Add(childChild)

						' Add a row to the table
						Dim col As Integer = 2
						If Not isStatic Then
							col = 1
						End If

						If eventsRow = -1 Then
							eventsRow = table.AddRow()
							table(0, eventsRow).ColumnSpan = 3
							table(0, eventsRow).Text = "Events"
							table(0, eventsRow).TextAlignment = TextAlignment.Left
							table(0, eventsRow).TextVerticalAlignment = AlignmentY.Center
							table(0, eventsRow).FontWeight = FontWeights.Bold
							table.Rows(eventsRow).Header = True
							table.Rows(eventsRow).Expanded = True
						End If

						Dim description As String = ""
						For Each attribute As Attribute In eventInfo.GetCustomAttributes(False)
							If TypeOf attribute Is DescriptionAttribute Then
								description = TryCast(attribute, DescriptionAttribute).Description
								Exit For
							End If
						Next

						Dim row As Integer = table.AddRow()

						If Not isStatic Then
							table(col, row).ColumnSpan = 2
						End If

						table(0, row).Image = DirectCast(Application.Current.Resources("Event"), ImageSource)
						table(0, row).ImageAlign = ImageAlign.Center
						table(col, row).Text = eventInfo.Name
						table(col, row).TextAlignment = TextAlignment.Left
						table(col, row).TextVerticalAlignment = AlignmentY.Center
						table(col, row).ToolTip = eventInfo.Name

						If isStatic Then
							table(1, row).Image = DirectCast(Application.Current.Resources("Static"), ImageSource)
							table(1, row).ImageAlign = ImageAlign.Center
						End If
					Next

					' Enumerate fields
					If Not type__1.IsInterface Then
						Dim fields As FieldInfo() = type__1.GetFields(BindingFlags.Instance Or BindingFlags.[Public] Or BindingFlags.NonPublic Or BindingFlags.[Static] Or BindingFlags.DeclaredOnly)

						Dim sortedFields As New ArrayList(fields)
						sortedFields.Sort(New MemberInfoComparer())

						Dim fieldsRow As Integer = -1

						For Each field As FieldInfo In sortedFields
							If Not field.IsPublic AndAlso Not field.IsFamily Then
								Continue For
							End If

							If field.IsSpecialName Then
								Continue For
							End If

							Dim fieldIcon As Integer = If((field.IsPublic), ImageField, ImagePField)

							Dim childChild = New System.Windows.Controls.TreeViewItem()
							'childChild.Header = field.Name;
							'childChild.Tag = field;
							childChild.Header = New Entity() With { _
							 .Name = field.Name, _
							 .Info = field, _
							 .Image = If((field.IsPublic), _ImageField, _ImagePField), _
							 .Type = If((field.IsPublic), ImageField, ImagePField) _
							}
							child.Items.Add(childChild)

							' Add a row to the table
							Dim col As Integer = 2
							If Not field.IsStatic Then
								col = 1
							End If

							If fieldsRow = -1 Then
								fieldsRow = table.AddRow()
								table(0, fieldsRow).ColumnSpan = 3
								table(0, fieldsRow).Text = "Fields"
								table(0, fieldsRow).TextAlignment = TextAlignment.Left
								table(0, fieldsRow).TextVerticalAlignment = AlignmentY.Center
								table(0, fieldsRow).FontWeight = FontWeights.Bold
								table.Rows(fieldsRow).Header = True
								table.Rows(fieldsRow).Expanded = True
							End If

							Dim row As Integer = table.AddRow()

							If Not field.IsStatic Then
								table(col, row).ColumnSpan = 2
							End If

							table(0, row).Image = DirectCast(Application.Current.Resources("Field"), ImageSource)
							table(0, row).ImageAlign = ImageAlign.Center
							table(col, row).Text = field.Name + " : " + field.FieldType.Name
							table(col, row).TextAlignment = TextAlignment.Left
							table(col, row).TextVerticalAlignment = AlignmentY.Center
							table(col, row).ToolTip = field.Name + " : " + field.FieldType.[Namespace] + Type.Delimiter + field.FieldType.Name
							table(col, row).Text = field.Name
							table(col, row).ToolTip = field.Name

							If field.IsStatic Then
								table(1, row).Image = DirectCast(Application.Current.Resources("Static"), ImageSource)
								table(1, row).ImageAlign = ImageAlign.Center
							End If
						Next
					End If
				End If
				Exit Select

			Case ImageEnum
				If True Then
					Dim fields As FieldInfo() = type__1.GetFields(BindingFlags.Instance Or BindingFlags.[Public] Or BindingFlags.NonPublic Or BindingFlags.[Static] Or BindingFlags.DeclaredOnly)

					Dim sortedFields As New ArrayList(fields)
					sortedFields.Sort(New MemberInfoComparer())

					For Each field As FieldInfo In sortedFields
						' We skip private stuff
						If field.IsPrivate Then
							Continue For
						End If

						' We skip internal stuff
						If field.IsAssembly Then
							Continue For
						End If

						' Skip __value field
						If field.IsSpecialName Then
							Continue For
						End If

						Dim childChild = New System.Windows.Controls.TreeViewItem()
						'childChild.Header = field.Name;
						'childChild.Tag = field;
						childChild.Header = New Entity() With { _
						 .Name = field.Name, _
						 .Info = field, _
						 .Image = _ImageEnumField, _
						 .Type = ImageEnumField _
						}
						child.Items.Add(childChild)

						' Add a row to the table
						Dim row As Integer = table.AddRow()

						table(1, row).Image = DirectCast(Application.Current.Resources("EnumField"), ImageSource)
						table(1, row).ImageAlign = ImageAlign.Center
						table(2, row).Text = field.Name
						table(2, row).TextAlignment = TextAlignment.Left
						table(2, row).TextVerticalAlignment = AlignmentY.Center
                        table(2, row).ToolTip = field.Name + " (" & System.Convert.ToInt64(field.GetValue(Nothing)) & ")"
					Next
				End If
				Exit Select
		End Select

		Return child
	End Function


	''' <summary>
	''' Gets the count of non-collapsed row.
	''' </summary>
	Private Function GetRowCount(ByVal t As TableNode) As Integer
		Dim rowCount As Integer = 0

		Dim inCollapse As Boolean = False
		For r As Integer = 0 To t.RowCount - 1
			If Not inCollapse Then
				rowCount += 1
			End If

			If t.Rows(r).Header Then
				If Not t.Rows(r).Expanded Then
					inCollapse = True
				Else
					inCollapse = False
				End If
			End If
		Next

		Return rowCount
	End Function

	Private Sub diagram_DrawNode(ByVal sender As Object, ByVal e As DrawNodeEventArgs)
		Dim table As TableNode = TryCast(e.Node, TableNode)
		If table Is Nothing Then
			Return
		End If

		Dim g As DrawingContext = e.Graphics

		'ImageSource image = _pureImages[(table.Tag as TreeNode).ImageIndex];
		'ImageSource image = (ImageSource)Application.Current.Resources["Event"];
		Dim image As ImageSource = TryCast(table.Tag, Entity).Image
		g.DrawImage(image, New Rect(2 * coef, 2, image.Width, image.Height))

		' Draw scroll arrows
		Dim arrow As ImageSource = DirectCast(Application.Current.Resources("ArrowUp"), ImageSource)
		Dim dest As New Point(table.Bounds.Width - arrow.Width - 2 * coef, 2)
		If table.CurrentScrollRow > 0 Then
			g.DrawImage(arrow, New Rect(dest.X, dest.Y, arrow.Width, arrow.Height))
		End If

		arrow = DirectCast(Application.Current.Resources("ArrowDown"), ImageSource)
		dest.Y += arrow.Height + 1

		Dim visibleRows As Integer = CInt((table.Bounds.Height - 7) / table.RowHeight)
		If table.CurrentScrollRow + visibleRows < GetRowCount(table) Then
			g.DrawImage(arrow, New Rect(dest.X, dest.Y, arrow.Width, arrow.Height))
		End If
	End Sub

	Private Sub diagram_NodeClicked(ByVal sender As Object, ByVal e As NodeEventArgs)
		Dim table As TableNode = TryCast(e.Node, TableNode)
		If table Is Nothing Then
			Return
		End If

		' Check scroll arrows
		Dim bounds As Rect = table.Bounds
		Dim mousePos As Point = e.MousePosition
		mousePos.X -= bounds.X
		mousePos.Y -= bounds.Y
		bounds.Offset(-bounds.X, -bounds.Y)

		Dim arrow As ImageSource = DirectCast(Application.Current.Resources("ArrowUp"), ImageSource)
		Dim imageBounds As New Rect(0, 0, arrow.Width, arrow.Height)


		Dim dest As New Point(bounds.Width - imageBounds.Width - 2, bounds.Y + 2)

		If table.CurrentScrollRow > 0 Then
			If New Rect(dest.X, dest.Y, imageBounds.Width, imageBounds.Height).Contains(mousePos) Then
				table.CurrentScrollRow -= 1
			End If
		End If

		dest.Y += imageBounds.Height + 1

		Dim visibleRows As Integer = CInt((bounds.Height - 7) / table.RowHeight)
		If table.CurrentScrollRow + visibleRows < GetRowCount(table) Then
			If New Rect(dest.X, dest.Y, imageBounds.Width, imageBounds.Height).Contains(mousePos) Then
				table.CurrentScrollRow += 1
			End If
		End If
	End Sub


	Private DefaultComparer As New Comparer(CultureInfo.InvariantCulture)

	Private ClassBrush As Brush = New LinearGradientBrush(Colors.White, Colors.LightBlue, 45)

	Private StructBrush As Brush = New LinearGradientBrush(Colors.White, Colors.LightPink, 45)

	Private InterfaceBrush As Brush = New LinearGradientBrush(Colors.White, Colors.LightGray, 45)

	Private EnumBrush As Brush = New LinearGradientBrush(Colors.White, Color.FromArgb(255, 255, 255, 160), 45)

	Private DelegateBrush As Brush = New LinearGradientBrush(Colors.White, Colors.PaleGreen, 45)

	Private Const ImageClass As Integer = 0
	Private Const ImageDelegate As Integer = 1
	Private Const ImageEnum As Integer = 2
	Private Const ImageEnumField As Integer = 3
	Private Const ImageEvent As Integer = 4
	Private Const ImageField As Integer = 5
	Private Const ImageInterface As Integer = 6
	Private Const ImageMethod As Integer = 7
	Private Const ImagePMethod As Integer = 8
	Private Const ImagePProperty As Integer = 9
	Private Const ImageProperty As Integer = 10
	Private Const ImageStatic As Integer = 11
	Private Const ImageStruct As Integer = 12
	Private Const ImageNamespace As Integer = 13
	Private Const ImagePField As Integer = 14

	Private _ImageClass As ImageSource = TryCast(Application.Current.Resources("Class"), ImageSource)
	Private _ImageDelegate As ImageSource = TryCast(Application.Current.Resources("Delegate"), ImageSource)
	Private _ImageEnum As ImageSource = TryCast(Application.Current.Resources("Enum"), ImageSource)
	Private _ImageEnumField As ImageSource = TryCast(Application.Current.Resources("EnumField"), ImageSource)
	Private _ImageEvent As ImageSource = TryCast(Application.Current.Resources("Event"), ImageSource)
	Private _ImageField As ImageSource = TryCast(Application.Current.Resources("ClasField"), ImageSource)
	Private _ImageInterface As ImageSource = TryCast(Application.Current.Resources("Interface"), ImageSource)
	Private _ImageMethod As ImageSource = TryCast(Application.Current.Resources("Method"), ImageSource)
	Private _ImagePMethod As ImageSource = TryCast(Application.Current.Resources("PMethod"), ImageSource)
	Private _ImagePProperty As ImageSource = TryCast(Application.Current.Resources("PProperty"), ImageSource)
	Private _ImageProperty As ImageSource = TryCast(Application.Current.Resources("Property"), ImageSource)
	Private _ImageStatic As ImageSource = TryCast(Application.Current.Resources("Static"), ImageSource)
	Private _ImageStruct As ImageSource = TryCast(Application.Current.Resources("Struct"), ImageSource)
	Private _ImageNamespace As ImageSource = TryCast(Application.Current.Resources("Namespace"), ImageSource)
	Private _ImagePField As ImageSource = TryCast(Application.Current.Resources("PField"), ImageSource)

	Private _openFileDialog As OpenFileDialog
	Private _workMap As Hashtable

	Private LeftCenterFormat As New StringFormat() With { _
	 .Alignment = StringAlignment.Near, _
	 .LineAlignment = StringAlignment.Center _
	}

End Class

Class Entity

	Public Property Name() As String
		Get
			Return m_Name
		End Get
		Set(ByVal value As String)
			m_Name = value
		End Set
	End Property

	Public Property Image() As ImageSource
		Get
			Return m_Image
		End Get
		Set(ByVal value As ImageSource)
			m_Image = value
		End Set
	End Property

	Public Property Info() As MemberInfo
		Get
			Return m_Info
		End Get
		Set(ByVal value As MemberInfo)
			m_Info = value
		End Set
	End Property

	Public Property Type() As Integer
		Get
			Return m_Type
		End Get
		Set(ByVal value As Integer)
			m_Type = value
		End Set
	End Property


	Private m_Name As String
	Private m_Image As ImageSource
	Private m_Info As MemberInfo
	Private m_Type As Integer

End Class